#include <stdio.h>
#include <locale.h>
#include <string.h>
#include <stdlib.h>

/************************************************************************/
/* PROGRAM:   fix_utf8.c      					*/
/* TASK:   this program makes some good guesses about what         */
/*   	any non-utf-8 chars from stdin are and prints           */
/*   	them to stdout                   			*/
/* NOTE:   the char datatype is used here to represent a byte	*/
/*   								*/
/* Julie L. Baumler          	     		November 20, 2012       */
/************************************************************************/

void print_mbchar( char *s) {
	size_t len = strlen(s);
	wchar_t wc;
	
	/* initialize */
	mbtowc(NULL, NULL, 0);
	int mb_len = mbtowc(&wc,s, len);
	printf("%lc", wc);
}

void main()
{
   int c;
   unsigned char convert[4];
   wchar_t as_wchar[2];

   /* counters */
   char i = 0;
   char cnt = 0;

   /* keep track of where we are in utf-8 data */
   char in_multibyte = 0;
   char bytes_left = 0;
   char output_now = 1;

   /* keep track of whether or not we are looking at a data string */
   char in_data = 0;

   setlocale(LC_CTYPE, "");


   /* read each character in input */
   while ((c = getchar()) != EOF )
   {
/*
if ( in_multibyte == 1 ){
printf("\nin_multibyte = %d, bytes_left = %d,  i = %d, c = %#x\n", 
	in_multibyte, bytes_left, i, c);
}
*/

/* VALID UTF-8
 * 00 to 7f hex (0 to 127): first and only byte of a sequence.
 * 80 to BF hex (128 to 191): continuing byte in a multi-byte sequence.
 * C2 to DF hex (194 to 223): first byte of a two-byte sequence.
 * E0 to EF hex (224 to 239): first byte of a three-byte sequence.
*/

/* VALID EUC-KR 
 * 2 bytes
 * A1A0 - FEFE
 */
                
   	if ( c >= (unsigned char)194 && c <= (unsigned char)223 && bytes_left == 0 )
   	{
   		/* first byte of a two-byte sequence. */
   		in_multibyte = 1;
   		bytes_left = 1;
   		convert[0] = c;
   		i=1;
   	}
   	else if ( c >= (unsigned char)224 && c<= (unsigned char)239 
   			&& bytes_left == 0 )
   	{
    		/* E0 to EF hex (224 to 239) */
   		/* first byte of a three-byte sequence. */
   		in_multibyte = 1;
   		bytes_left = 2;
   		convert[0] = c;
   		i=1;
   	}
   	else if ( in_multibyte && 
   		(c >= (unsigned char)128 
   			&& c<= (unsigned char)191) )
   	{
  		/* continuing byte in a multi-byte sequence */
   		/* This is normal multi-byte character */
   		/* (at least so far)		       */
   		convert[i] = c;
   		bytes_left--;
   		i++;
/*
printf("\nbytes_left = %d,  i = %d, c = %#x\n", bytes_left, i, c);
*/
   		if ( bytes_left == 0 )
   		{
/*
printf("No bytes left!\n");
*/
   			/* we found a utf-8 character convert to wchar_t*/
			/* put a NULL byte at the end of the string */
			convert[i] = '\0';
			print_mbchar( &convert[0] );

/*
   			for( cnt = 0; cnt < i; cnt++ )
   			{
   				printf("%02hhx", convert[cnt]);
   			}
			printf("\n");
*/
   			/* reset flags and counters */
   			in_multibyte = 0;
   			i=0;
   		} 
   	}
   	else 
   	{
		/* EITHER */
		/*	 this is a single byte char */
                /* OR */
		/* 	we just discovered that we have a faux multibyte */
                /* either way, we need to convert whatever bytes we have */
                /* and output them */
   	
   		/* add this byte to our array of bytes to be converted */
   		convert[i] = c;
		i++;

   		for( cnt = 0; cnt < i; cnt++ )
		{
   
   			switch ( convert[cnt] ) 
   			{
				case (unsigned char)'\x80':
					/* non-printing control char 
					   or possibly monetary symbol
					   but the absence of the later
					   makes sense in context */
					break;
				case (unsigned char)'\x81':
					/* non-printing control char */
					break;
   				case (unsigned char)'\x83':
   					printf("&#pound;");
   					break;
   				case (unsigned char)'\x84':
   					printf("&#quot;");
   					break;
   				case (unsigned char)'\x85':
   					printf("...");
   					break;
   				case (unsigned char)'\x91':
   					printf("&#39;");
   					break;
   				case (unsigned char)'\x92':
   					printf("&#39;");
   					break;
   				case (unsigned char)'\x93':
   					printf("&quot;");
   					break;
   				case (unsigned char)'\x94':
   					printf("&quot;");
   					break;
   				case (unsigned char)'\x95':
   					break;
   				case (unsigned char)'\x96':
   					putchar('-');
   					break;
   				case (unsigned char)'\x97':
   					putchar('-');
   					break;
   				case (unsigned char)'\xa0':
   					putchar(' ');
   					break;
   				case (unsigned char)'\xa1':
   					printf("&iexcl;");
   					break;
   				case (unsigned char)'\xa2':
   					printf("&cent;");
   					break;
   				case (unsigned char)'\xa6':
   					putchar('|');
   					break;
   				case (unsigned char)'\xae':
   					printf("&reg;");
   					break;
   				case (unsigned char)'\xb0':
   					printf("&deg;");
   					break;
   				case (unsigned char)'\xb4':
   					printf("&acute;");
   					break;
   				case (unsigned char)'\xb7':
   					putchar('-');
   					break;
   				case (unsigned char)'\xba':
   					printf("&ordm;");
   					break;
   				case (unsigned char)'\xbd':
   					printf("1/2");
   					break;
   				case (unsigned char)'\xbc':
   					printf("1/4");
   					break;
   				case (unsigned char)'\xbf':
   					printf("&iquest;");
   					break;
   				case (unsigned char)'\xc1':
   					printf("&Aacute;");
   					break;
   				case (unsigned char)'\xc3':
   					printf("&Atilde;");
   					break;
   				case (unsigned char)'\xd1':
   					printf("&Ntilde;");
   					break;
   				case (unsigned char)'\xd3':
   					printf("&szlig;");
   					break;
   				case (unsigned char)'\xe0':
   					printf("&agrave;");
   					break;
   				case (unsigned char)'\xe1':
   					printf("&aacute;");
   					break;
   				case (unsigned char)'\xe3':
   					printf("&atilde;");
   					break;
   				case (unsigned char)'\xe7':
   					printf("&ccedil;");
   					break;
   				case (unsigned char)'\xe8':
   					printf("&egrave;");
   					break;
   				case (unsigned char)'\xe9':
   					printf("&eacute;");
   					break;
   				case (unsigned char)'\xea':
   					printf("&ecirc;");
   					break;
   				case (unsigned char)'\xed':
   					printf("&iacute;");
   					break;
   				case (unsigned char)'\xf0':
   					printf("&eth;");
   					break;
   				case (unsigned char)'\xf1':
   					printf("&ntilde;");
   					break;
   				case (unsigned char)'\xf2':
   					printf("&ograve;");
   					break;
   				case (unsigned char)'\xf3':
   					printf("&oacute;");
   					break;
   				case (unsigned char)'\xf4':
   					printf("&ocirc;");
   					break;
   				case (unsigned char)'\xf5':
   					printf("&otilde;");
   					break;
   				case (unsigned char)'\xf6':
   					printf("&ouml;");
   					break;
   				case (unsigned char)'\xf7':
   					printf("&divide;");
   					break;
   				case (unsigned char)'\xf8':
   					printf("&oslash;");
   					break;
   				case (unsigned char)'\xf9':
   					printf("&ugrave;");
   					break;
   				case (unsigned char)'\xfa':
   					printf("&uacute;");
   					break;
   				case (unsigned char)'\xfb':
   					printf("&ucirc;");
   					break;
   				case (unsigned char)'\xfc':
   					printf("&uuml;");
   					break;
   				case (unsigned char)'\xfd':
   					printf("&yacute;");
   					break;
   				case (unsigned char)'\xfe':
   					printf("&thorn;");
   					break;
   				default:
   					/* print char */
   					putchar(c);	
   			}
   		}

   		/* reset counters and flags */
   		/* they may or may not have these values already */
   		bytes_left = 0;
   		in_multibyte = 0;
   		i = 0;
   		
   	}
   }


   /* quit when EOF is reached */
}
